from sklearn import datasets, preprocessing
import pdb
import numpy as np
import math
import scipy.io as sio
import os
import math
from scipy.sparse import csr_matrix
import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import SparseTensorValue as STV
import xclib.evaluation.xc_metrics as xc_metrics
import time
import h5py
import ctypes
from numpy.ctypeslib import ndpointer

MASK_MIN = -np.inf

class Y(object):
    def __init__(self, m, n, nnzs, dtype, nonzero=None):
        self.m = m
        self.n = n
        self.nnzs = nnzs
        self.dtype = dtype

        self.tpk_idx = None
        self.tpk_mask = None
        self.pl_mask = None

        if nonzero is not None:
            nonzero = np.array(nonzero)

            self.Label = csr_matrix((np.ones(nnzs, dtype=dtype), nonzero), shape=(m, n))
            self.Weight = csr_matrix((np.ones(nnzs, dtype=dtype), nonzero), shape=(m, n))

            self.Sigma = csr_matrix((np.ones(nnzs, dtype=dtype), nonzero), shape=(m, n))
            self.Y_hat = csr_matrix((np.ones(nnzs, dtype=dtype), nonzero), shape=(m, n))
            self.Y_imp = csr_matrix((np.ones(nnzs, dtype=dtype), nonzero), shape=(m, n))

    def transpose(self):
        yT = Y(self.n, self.m, self.nnzs, self.dtype)
        nonzero = self.Label.transpose().nonzero()

        def gen_T(array):
            return csr_matrix((np.array(array.transpose()[nonzero]).squeeze(), nonzero), shape=(self.n, self.m))

        yT.Label = gen_T(self.Label)
        yT.Weight = gen_T(self.Weight)
        yT.Sigma = gen_T(self.Sigma)
        yT.Y_hat = gen_T(self.Y_hat)
        yT.Y_imp = gen_T(self.Y_imp)
        return yT

    def read_weights(self, filename, dtype, is_value_based):
        nzinfos = []
        with open(filename, 'r') as fin:
            for line in fin:
                line = line.strip()
                nzinfos.extend([(float(_w.split(':')[0]), float(_w.split(':')[1])) \
                        for _w in line.split(',')])
        rewards, weights = zip(*nzinfos)
        if is_value_based:
            self.Label.data = 2.0*(np.array(rewards, dtype=dtype) == 1).astype(dtype) - 1.0
            self.Weight.data = 1./np.array(weights, dtype=dtype)
        else:
            self.Weight.data = np.array(rewards, dtype=dtype)/np.array(weights, dtype=dtype)

    def read_tpk(self, filename, tpk_num, va_flag, dtype):
        self.tpk_idx = np.loadtxt(filename, dtype=np.int64, delimiter=',')[:, :tpk_num]
        if not va_flag:
            self.tpk_mask = np.vstack([np.isin(self.tpk_idx[i, :], [y], assume_unique=True) for i, y in enumerate(self.Label.nonzero()[1])]).astype(dtype)

    def read_pl_mask(self, filename, loss_type):
        self.pl_mask = gen_skip_mask(self.m, self.n, filename)
        if self.tpk_idx is not None and loss_type == 4:
            self.pl_mask = np.take_along_axis(self.pl_mask, self.tpk_idx, axis=1)

class ConfigClass(object):
    def __init__(self, args, Ytr, skip_mask, R):
        super(ConfigClass, self).__init__()
        self.args = args
        self.iter_max = args.iter_max
        self.skip_mask = skip_mask

        self.seed = args.seed

        if self.seed is None:
            print('You choose not to specify a random seed.'+\
                'A different result is produced after each run.')
        elif isinstance(self.seed, int) and self.seed >= 0:
            print('You specify random seed {}.'.format(self.seed))
        else:
            raise ValueError('Only accept None type or nonnegative integers for'+\
                    ' random seed argument!')

        self.u_set = args.u_set
        self.v_set = args.v_set
        self.val_set = args.val_set

        self.m = int(Ytr.m)
        self.n = int(Ytr.n)

        self.nnzs = int(Ytr.nnzs)
        self.R = R + args.zeta


        self.r = args.r
        self.zeta = args.zeta
        self.tpk_num = args.tpk_num
        self.slot_num = args.slot_num

        self.GNsize = min(args.GNsize, self.m, self.n)
        self.C = args.C
        self.net_u = args.net_u
        self.net_v = args.net_v

        self.k = args.k
        self.k1 = args.k1

        self.xi = args.xi
        self.CGmax = args.CGmax
        self.omega=args.omega
        self._lambda = args._lambda
        self.eta = args.eta
        self.lr = args.lr
        self.alpha = args.alpha
        self.from_prev = args.from_prev
        self.thres = args.thres
        self.b_ratio = args.b_ratio
        #ips:0, log-ips:1, bce:2, bce_sq:3
        self.loss_type = args.loss_type
        print("loss_type:",self.loss_type)

        self.bsize = args.bsize

        self.sogram_bsize = max(round(self.b_ratio*self.nnzs), 64)

        self.bsize_i = round(self.m*(self.b_ratio**0.5))
        #self.bsize_j = round(self.n*(self.b_ratio**0.5))
        #FIXME
        self.bsize_j = self.n

        self.gramian_alpha = 1 if self.sogram_bsize == self.nnzs else args.gramian_alpha

        self.optim = args.optim
        if self.optim not in ('SGD', 'NewtonCG', 'Adam', 'Sogram', 'Minibatch', 'General'):
            raise ValueError('Only support SGD, Adam, NewtonCG, Sogram, and Minibatch optimizer!')

        self.log_file = args.log_file
        self.model_file = args.model_file
        self.screen_log_only = args.screen_log_only
        self.profile = args.profile

        precision = 'float' + str(args.precision)
        self.dtype = getattr(tf, precision)
        self.dtype_np = getattr(np, precision)

        if self.screen_log_only:
            print('You choose not to store running log. Only store model to {}'.format(self.log_file))
        else:
            print('Saving log to: {}'.format(self.log_file))
            dir_name, _ = os.path.split(self.log_file)
            if dir_name and not os.path.isdir(dir_name):
                os.makedirs(dir_name, exist_ok=True)

        dir_name, _ = os.path.split(self.model_file)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name, exist_ok=True)
        self.elapsed_time = 0.0

def gen_label_matrix(labels_list, max_label_idx, dtype):
    m  = len(labels_list)
    n = -1

    rows = []
    columns = []

    nnz = 0
    row_idx = 0

    for labels in labels_list:
        for label in labels:
            rows.append(int(row_idx))
            columns.append(int(label))
        this_nnzs = len(labels)
        nnz += this_nnzs
        n = max(n, max(labels))
        row_idx += 1
    return Y(int(m), max(int(n+1), max_label_idx), nnz, dtype, [rows, columns])

class Data():
    def __init__(self, path, dtype, **kwargs):
        self.read_data(path, dtype, **kwargs)

    def __getattr__(self, attr):
        return getattr(self.feat, attr)

    def __getitem__(self, idx):
        sub = self.__class__(None, self.dtype)
        sub.feat = self.feat[idx]
        # sub.label = self.label[idx]
        return sub

    def read_data(self, path, **kwargs):
        raise NotImplementedError

    def transform(self):
        raise NotImplementedError

    def microbatches(self, config, bsize):
        m = self.feat.shape[0]
        num_segment = math.ceil(m/bsize)
        for i in range(num_segment):
            load_time = time.time()
            idx = np.arange(i * bsize, min((i+1) * bsize, m))
            start, end = idx[0], idx[-1]+1
            subset = slice(start, end)
            batch_input = self.transform(self.feat[subset])
            config.elapsed_time += time.time() - load_time
            yield subset, batch_input

class SparseData(Data):
    def __init__(self, path, dtype, **kwargs):
        super().__init__(path, dtype, **kwargs)

    def read_data(self, path, dtype, max_feat_idx=-1, max_label_idx=-1, feat=None, has_label=True):
        if path is None:
            self.feat = None
            self.label = None
        elif feat is not None:
            self.feat = feat
        else:
            X, y = datasets.load_svmlight_file(path, zero_based=True, multilabel=True)
            X = csr_matrix((X.data, X.nonzero()), shape=(X.shape[0], max(X.shape[1], max_feat_idx)))
            if max_feat_idx > 0:
                X = X[:, :max_feat_idx]
            X = preprocessing.normalize(X, norm="l2")
            self.feat = X
            if has_label:
                self.label = gen_label_matrix(y, max_label_idx, dtype)

    def transform(self, B):
        return STV(np.array(B.nonzero()).transpose(), B.data, B.shape)

class DenseData(Data):
    def __init__(self, path, dtype, **kwargs):
        super().__init__(path, dtype, **kwargs)

    def read_data(self, path, dtype, feat=None):
        if path is None:
            self.feat = None
            self.label = None
        elif feat is not None:
            self.feat = feat
        else:
            data = h5py.File(path, 'r')
            self.feat = data['images']

            m = data['m'][()]
            nnz = data['nnz'][()]
            rows = data['rows'][()]
            cols = data['cols'][()]
            self.label = Y(m, m, nnz, dtype, [rows, cols])

    def transform(self, B):
        return B

def read_data(loader_type, u_path, v_path, val_path, dtype):
    if loader_type == 'sparse':
        trainR = SparseData(v_path, dtype, has_label=False)
        trainL = SparseData(u_path, dtype, max_label_idx=trainR.feat.shape[0])

        validL = SparseData(val_path, dtype,
            max_feat_idx=trainL.feat.shape[1],
            max_label_idx=trainL.label.n)
        validR = SparseData(val_path, dtype, feat=trainR.feat)
    elif loader_type == 'dense':
        trainL = DenseData(u_path, dtype)
        trainR = DenseData(u_path, dtype, feat=trainL.feat)

        validL = DenseData(val_path, dtype)
        validR = DenseData(val_path, dtype, feat=validL.feat)
    else:
        raise ValueError("No such loader_type: %s"%loader_type)

    return trainL, trainR, validL, validR


def _cal_output(sess, input, input_placeholder, output, bsize, config, dtype):
    m = input.feat.shape[0]
    output_mat = np.zeros((m, output.shape[1].value), dtype=dtype)
    for subset, batch_input in input.microbatches(config, bsize):
        output_mat[subset] = sess.run(output, feed_dict={input_placeholder:batch_input})
    return output_mat

def gen_skip_mask(m, n, skip_mask=None):
    if skip_mask is None:
        return csr_matrix((np.ones(0, dtype=int), ([], [])), shape=(m, n))
    skip_mask_file = open(skip_mask, 'r')
    lines = skip_mask_file.readlines()
    i_array, j_array = [], []
    i,nnzs = 0, 0
    for l in lines:
        js = l.strip().split(',')
        for j in js:
            if not j:
                continue
            nnzs += 1
            i_array.append(i)
            j_array.append(int(j))
        i = i+1
    nonzeros = (np.array(i_array), np.array(j_array))
    return csr_matrix((np.ones(nnzs, dtype=int), nonzeros), shape=(i, n))

def save_PQ(sess, network, validL, validR, config, dtype):
    Yte = validL.label.Label
    u,v,p,q = network
    bsize = 2**17

    P = _cal_output(sess, validL, u, p, bsize, config, dtype)
    Q = _cal_output(sess, validR, v, q, bsize, config, dtype)

    save_dir = os.path.dirname(config.model_file)
    set_name = os.path.basename(config.args.val_set)[:-4]
    fp = os.path.join(save_dir, 'P_%s.npy'%set_name)
    fq = os.path.join(save_dir, 'Q_%s.npy'%set_name)
    np.save(fp, P)
    np.save(fq, Q)

    return
 
def predict(sess, network, validL, validR, config, dtype, k=10):
    Yte = validL.label.Label
    tpk_idx = validL.label.tpk_idx
    u,v,p,q = network
    bsize, skip_mask = 2**17, config.skip_mask

    P = _cal_output(sess, validL, u, p, bsize, config, dtype)
    Q = _cal_output(sess, validR, v, q, bsize, config, dtype)
    
    # score_mat = P.dot(Q.T)
    m, n = P.shape[0], Q.shape[0]
    segment_m, segment_n = math.ceil(m/bsize), math.ceil(n/bsize)
    precision, ndcg = np.zeros(k,dtype=P.dtype), np.zeros(k,dtype=P.dtype)
    precision_tpk, ndcg_tpk = np.zeros(k,dtype=P.dtype), np.zeros(k,dtype=P.dtype)

    for i in range(segment_m):
        i_start, i_end = i*bsize, min((i+1)*bsize, m)
        score_mat = np.zeros((i_end-i_start, n), dtype=P.dtype)
        for j in range(segment_n):
            j_start, j_end = j*bsize, min((j+1)*bsize, n)
            score_mat[:, j_start:j_end] = P[i_start:i_end].dot(Q[j_start:j_end].T)
        if skip_mask is not None:
            score_mat[skip_mask[i_start:i_end].nonzero()] = MASK_MIN
        _precision = xc_metrics.precision(score_mat, Yte[i_start:i_end], k=k)
        _ndcg = xc_metrics.ndcg(score_mat, Yte[i_start:i_end], k=k)
        if tpk_idx is not None:
            score_mat_tpk = np.take_along_axis(score_mat, tpk_idx[i_start:i_end], axis=1)
            Yte_tpk = np.take_along_axis(Yte[i_start:i_end], tpk_idx[i_start:i_end], axis=1)
            _precision_tpk = xc_metrics.precision(score_mat_tpk, Yte_tpk, k=k)
            _ndcg_tpk = xc_metrics.ndcg(score_mat_tpk, Yte_tpk, k=k)
        else:
            _precision_tpk = 0.0
            _ndcg_tpk = 0.0
        precision += (i_end-i_start)*_precision
        ndcg += (i_end-i_start)*_ndcg
        precision_tpk += (i_end-i_start)*_precision_tpk
        ndcg_tpk += (i_end-i_start)*_ndcg_tpk
    precision /= m
    ndcg /= m
    precision_tpk /= m
    ndcg_tpk /= m
    return precision, ndcg, precision_tpk, ndcg_tpk

def vectorize(tensors):
    if isinstance(tensors, list) or isinstance(tensors, tuple):
        vector = [tf.reshape(tensor, [-1]) for tensor in tensors]
        return tf.concat(vector, 0)
    else:
        return tensors

def inverse_vectorize(vector, param):
    if isinstance(vector, list):
        return vector
    else:
        tensors = []
        offset = 0
        num_total_param = np.sum([np.prod(p.shape.as_list()) for p in param])
        for p in param:
            numel = np.prod(p.shape.as_list())
            tensors.append(tf.reshape(vector[offset: offset+numel], p.shape))
            offset += numel

        assert offset == num_total_param
        return tensors

def calc_norm(v):
    # default: frobenius norm
    if isinstance(v, list):
        norm = 0.
        for p in v:
            norm = norm + tf.norm(tensor=p)**2
        return norm**0.5
    else:
        return tf.norm(tensor=v)

def zero_tensor(g):
    return tf.assign(g, tf.zeros_like(g))

def get_cfuncs(dtype):
    if dtype == tf.float64:
        dtype_c = ctypes.c_double
    elif dtype == tf.float32:
        dtype_c = ctypes.c_float
    else:
        print('dtype not supported!')
        exit(1)

    lib = ctypes.cdll.LoadLibrary("./c_funs/cfuns.so")

    cBQ = lib.cBQ
    cBQ.restype = None
    cBQ.argtypes = [
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_int,
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ctypes.c_size_t,
        ctypes.c_size_t,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_double,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
    ]

    cz = lib.cz
    cz.restype = None
    cz.argtypes = [
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_int,
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ctypes.c_size_t,
        ctypes.c_size_t,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_double,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
    ]

    cpred_val = lib.cpred_val
    cpred_val.restype = ctypes.c_double
    cpred_val.argtypes = [
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_int,
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ctypes.c_size_t,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_double,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
    ]

    cpred_val_sogram = lib.cpred_val_sogram
    cpred_val_sogram.restype = ctypes.c_double
    cpred_val_sogram.argtypes = [
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_int,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ctypes.c_int,
        ctypes.c_size_t,
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
        ctypes.c_double,
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
        ndpointer(dtype_c, flags="C_CONTIGUOUS"),
    ]

    return cBQ, cz, cpred_val, cpred_val_sogram

